import json
from uuid import uuid4

from ligo.data_model.receptor.Receptor import Receptor
from ligo.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence


class BCKReceptor(Receptor):
    FIELDS = {'heavy': object, 'kappa': object, 'identifier': str, 'metadata': dict, 'version': str}
    version = "1"

    @classmethod
    def create_from_record(cls, record):
        if 'version' in record.dtype.names and record['version'] == BCKReceptor.version:

            heavy_record = record[['heavy_' + name for name in ReceptorSequence.get_record_names()]]
            heavy_record.dtype.names = ReceptorSequence.get_record_names()

            kappa_record = record[['kappa_' + name for name in ReceptorSequence.get_record_names()]]
            kappa_record.dtype.names = ReceptorSequence.get_record_names()

            metadata = json.loads(record['metadata']) if record['metadata'] != '' else None

            return BCKReceptor(heavy=ReceptorSequence.create_from_record(heavy_record),
                               kappa=ReceptorSequence.create_from_record(kappa_record),
                               identifier=record['identifier'], metadata=metadata)
        else:
            raise NotImplementedError(f"Supported ({BCKReceptor.version}) and available version differ, but there is no converter available.")

    @classmethod
    def get_record_names(cls):
        return ['heavy_' + name for name in ReceptorSequence.get_record_names()] \
               + ['kappa_' + name for name in ReceptorSequence.get_record_names()] \
               + [name for name in cls.FIELDS if name not in ['heavy', 'kappa']]

    def __init__(self, heavy: ReceptorSequence = None, kappa: ReceptorSequence = None, metadata: dict = None, identifier: str = None):
        super().__init__(metadata=metadata, identifier=identifier)
        self.heavy = heavy
        self.kappa = kappa

    def get_chains(self):
        return ["heavy", "kappa"]



import copy
import logging
import math
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from ligo.data_model.dataset.ElementDataset import ElementDataset
from ligo.data_model.receptor.Receptor import Receptor


class ReceptorDataset(ElementDataset):
    """A dataset class for receptor datasets (paired chain). All the functionality is implemented in ElementDataset class, except creating a new
    dataset and obtaining metadata. """

    @classmethod
    def build(cls, **kwargs):
        return ReceptorDataset(**kwargs)

    @classmethod
    def build_from_objects(cls, receptors: List[Receptor], file_size: int, path: Path, name: str = None, labels: dict = None):

        file_count = math.ceil(len(receptors) / file_size)
        file_names = [path / f"batch{''.join(['0' for i in range(1, len(str(file_count)) - len(str(index)) + 1)])}{index}.npy"
                      for index in range(1, file_count + 1)]

        for index in range(file_count):
            receptor_matrix = np.core.records.fromrecords(
                [receptor.get_record() for receptor in receptors[index * file_size:(index + 1) * file_size]],
                names=type(receptors[0]).get_record_names())
            np.save(str(file_names[index]), receptor_matrix, allow_pickle=False)

        return ReceptorDataset(filenames=file_names, file_size=file_size, name=name, labels=labels,
                               element_class_name=type(receptors[0]).__name__ if len(receptors) > 0 else None)

    def get_metadata(self, field_names: list, return_df: bool = False):
        """Returns a dict or an equivalent pandas DataFrame with metadata information from Receptor objects for provided field names"""
        result = {field: [] for field in field_names}
        for receptor in self.get_data():
            for field in field_names:
                result[field].append(receptor.metadata[field] if receptor.metadata and field in receptor.metadata else None)

        for field in field_names:
            if all(item is None for item in result[field]):
                logging.warning(f"{ReceptorDataset.__name__}: none of the receptors in the dataset {self.name} have metadata field '{field}'. "
                                f"Returning 'None' instead...")
                result[field] = None

        return pd.DataFrame(result) if return_df else result

    def clone(self, keep_identifier: bool = False):
        dataset = ReceptorDataset(self.labels, copy.deepcopy(self.encoded_data), copy.deepcopy(self.filenames), file_size=self.file_size,
                                  name=self.name, element_class_name=self.element_generator.element_class_name)
        if keep_identifier:
            dataset.identifier = self.identifier
        dataset.element_ids = self.element_ids
        return dataset



import math
from pathlib import Path

import numpy as np

from ligo.util.ReflectionHandler import ReflectionHandler


class ElementGenerator:

    def __init__(self, file_list: list, file_size: int = 1000, element_class_name: str = ""):
        self.file_list = file_list
        self.file_lengths = [-1 for _ in range(len(file_list))]
        self.file_size = file_size
        self.element_class_name = element_class_name

    def _load_batch(self, current_file: int):

        element_class = ReflectionHandler.get_class_by_name(self.element_class_name, "data_model")
        assert hasattr(element_class, 'create_from_record'), \
            f"{ElementGenerator.__name__}: cannot load the binary file, the class {element_class.__name__} has no 'create_from_record' method."

        try:
            elements = [element_class.create_from_record(el) for el in np.load(self.file_list[current_file], allow_pickle=False)]
        except ValueError as error:
            raise ValueError(f'{ElementGenerator.__name__}: an error occurred while creating an object from binary file. Details: {error}')

        return elements

    def _get_element_count(self, file_index: int):

        # TODO: make this abstract and move implementation to specific generator: count elements in file for new format

        if self.file_lengths[file_index] == -1:
            with self.file_list[file_index].open("rb") as file:
                count = len(np.load(file))
            self.file_lengths[file_index] = count

        return self.file_lengths[file_index]

    def get_element_count(self):
        for index in range(len(self.file_list)):
            if self.file_lengths[index] == -1:
                self._get_element_count(index)
        return sum(self.file_lengths)

    def build_batch_generator(self):
        """
        creates a generator which will return one batch of elements at the time

        :param batch_size: how many elements should be returned at once (default 1)
        :return: element generator
        """

        for current_file_index in range(len(self.file_list)):
            batch = self._load_batch(current_file_index)
            yield batch

    def build_element_generator(self):
        """
        creates a generator which will return one element at the time
        :return: element generator
        """
        for current_file_index in range(len(self.file_list)):
            batch = self._load_batch(current_file_index)
            for element in batch:
                yield element

    def make_subset(self, example_indices: list, path: Path, dataset_type: str, dataset_identifier: str):
        if example_indices is None or len(example_indices) == 0:
            raise RuntimeError(f"{ElementGenerator.__name__}: no examples were specified to create the dataset subset. "
                               f"Dataset type was {dataset_type}, dataset identifier: {dataset_identifier}.")
        batch_size = self.file_size
        elements = []
        file_count = 1

        example_indices.sort()

        batch_filenames = self._prepare_batch_filenames(len(example_indices), path, dataset_type, dataset_identifier)

        for index, batch in enumerate(self.build_batch_generator()):
            extracted_elements = self._extract_elements_from_batch(index, batch_size, batch, example_indices)
            elements.extend(extracted_elements)

            if len(elements) >= self.file_size or len(elements) == len(example_indices):
                self._store_elements_to_file(batch_filenames[file_count - 1], elements[:self.file_size])
                file_count += 1
                elements = elements[self.file_size:]

        if len(elements) > 0:
            self._store_elements_to_file(batch_filenames[file_count - 1], elements)

        return batch_filenames

    def _prepare_batch_filenames(self, example_count: int, path: Path, dataset_type: str, dataset_identifier: str):
        batch_count = math.ceil(example_count / self.file_size)
        digits_count = len(str(batch_count)) + 1
        filenames = [path / f"{dataset_identifier}_{dataset_type}_batch{''.join(['0' for i in range(digits_count - len(str(index)))])}{index}.npy"
                     for index in range(batch_count)]
        return filenames

    def _store_elements_to_file(self, path, elements):
        if isinstance(elements, list) and len(elements) > 0:
            element_matrix = np.core.records.fromrecords([el.get_record() for el in elements], names=type(elements[0]).get_record_names())
            np.save(str(path), element_matrix, allow_pickle=False)

    def _extract_elements_from_batch(self, index, batch_size, batch, example_indices):
        upper_limit, lower_limit = (index + 1) * batch_size, index * batch_size
        batch_indices = [ind for ind in example_indices if lower_limit <= ind < upper_limit]

        assert len(batch_indices) == 0 or max(batch_indices) - lower_limit < len(
            batch), f"ElementGenerator: Found batch of size {len(batch)}, but expected {batch_size}. " \
                    f"Are the batch files sorted correctly? All files except the last file must have batch size {batch_size}."

        elements = [batch[i - lower_limit] for i in batch_indices]
        return elements

    def get_data_from_index_range(self, start_index: int, end_index: int):
        elements = []
        start_file_index = start_index // self.file_size
        end_file_index = min(len(self.file_list) - 1, end_index // self.file_size)
        for current_file_index in range(start_file_index, end_file_index + 1):
            batch = self._load_batch(current_file_index)
            i = 0 if current_file_index != start_file_index else start_index % self.file_size
            while len(elements) < end_index - start_index + 1 and i < len(batch):
                elements.append(batch[i])
                i += 1
        return elements



from inspect import signature
from pathlib import Path

from ligo.IO.dataset_import.DataImport import DataImport
from ligo.dsl.definition_parsers.DefinitionParserOutput import DefinitionParserOutput
from ligo.dsl.definition_parsers.MotifParser import MotifParser
from ligo.dsl.definition_parsers.SignalParser import SignalParser
from ligo.dsl.definition_parsers.SimulationParser import SimulationParser
from ligo.dsl.symbol_table.SymbolTable import SymbolTable
from ligo.simulation.SimConfig import SimConfig
from ligo.simulation.SimConfigItem import SimConfigItem
from ligo.simulation.generative_models.GenerativeModel import GenerativeModel
from ligo.simulation.implants.LigoPWM import LigoPWM
from ligo.simulation.implants.SeedMotif import SeedMotif
from ligo.simulation.implants.Signal import Signal
from ligo.util.PathBuilder import PathBuilder
from ligo.util.ReflectionHandler import ReflectionHandler
from scripts.DocumentatonFormat import DocumentationFormat
from scripts.specification_util import write_class_docs, make_docs


class DefinitionParser:

    @staticmethod
    def parse(workflow_specification: dict, symbol_table: SymbolTable, result_path: Path):

        specs = workflow_specification["definitions"]

        specs_defs = {}

        for parser in [MotifParser, SignalParser, SimulationParser]:
            symbol_table, new_specs = DefinitionParser._call_if_exists(parser.keyword, parser.parse, specs, symbol_table, result_path)
            specs_defs[parser.keyword] = new_specs

        return DefinitionParserOutput(symbol_table=symbol_table, specification=workflow_specification), specs_defs

    @staticmethod
    def _call_if_exists(key: str, method, specs: dict, symbol_table: SymbolTable, path=None):
        if key in specs:
            if "path" in signature(method).parameters:
                return method(specs[key], symbol_table, path)
            else:
                return method(specs[key], symbol_table)
        else:
            return symbol_table, {}

    @staticmethod
    def generate_docs(path: Path):
        def_path = PathBuilder.build(path / "definitions")
        DefinitionParser.make_dataset_docs(def_path)
        DefinitionParser.make_simulation_docs(def_path)
        DefinitionParser.make_gen_model_docs(def_path)

    @staticmethod
    def make_simulation_docs(path: Path):

        classes_to_document = [DocumentationFormat(SeedMotif, SeedMotif.__name__, DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(LigoPWM, "PWM", DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(Signal, Signal.__name__, DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(SimConfig, "Simulation config", DocumentationFormat.LEVELS[1]),
                               DocumentationFormat(SimConfigItem, "Simulation config item", DocumentationFormat.LEVELS[1])]

        file_path = path / "simulation.rst"
        with file_path.open("w") as file:
            for doc_format in classes_to_document:
                write_class_docs(doc_format, file)

    @staticmethod
    def make_dataset_docs(path: Path):
        import_classes = ReflectionHandler.all_nonabstract_subclasses(DataImport, "Import", "dataset_import/")
        make_docs(path, import_classes, "datasets.rst", "Import")

    @staticmethod
    def make_gen_model_docs(path: Path):
        gen_models = ReflectionHandler.all_nonabstract_subclasses(GenerativeModel, "", "simulation/generative_models/")
        make_docs(path, gen_models, "gen_models.rst", "")




from ligo.presentation.PresentationFormat import PresentationFormat
from ligo.presentation.html.FeasibilitySummaryHTMLBuilder import FeasibilitySummaryHTMLBuilder
from ligo.presentation.html.LIgOSimulationHTMLBuilder import LIgOSimulationHTMLBuilder
from ligo.simulation.LigoSimState import LigoSimState
from ligo.workflows.instructions.ligo_sim_feasibility.FeasibilitySummaryInstruction import FeasibilitySummaryState


class PresentationFactory:

    @staticmethod
    def make_presentation_builder(state, presentation_format: PresentationFormat):
        if isinstance(state, LigoSimState) and presentation_format == PresentationFormat.HTML:
            return LIgOSimulationHTMLBuilder
        elif isinstance(state, FeasibilitySummaryState) and presentation_format == PresentationFormat.HTML:
            return FeasibilitySummaryHTMLBuilder
        else:
            raise ValueError(f"PresentationFactory: state and format combination ({type(state).__name__}, {presentation_format.name}) "
                             f"is not supported.")



import logging
import os
from pathlib import Path

from ligo.dsl.DefaultParamsLoader import DefaultParamsLoader
from ligo.dsl.definition_parsers.DefinitionParserOutput import DefinitionParserOutput
from ligo.dsl.symbol_table.SymbolTable import SymbolTable
from ligo.dsl.symbol_table.SymbolType import SymbolType
from ligo.environment.EnvironmentSettings import EnvironmentSettings
from ligo.util.Logger import log
from ligo.util.ParameterValidator import ParameterValidator
from ligo.util.PathBuilder import PathBuilder
from ligo.util.ReflectionHandler import ReflectionHandler
from ligo.workflows.instructions.Instruction import Instruction
from scripts.DocumentatonFormat import DocumentationFormat
from scripts.specification_util import write_class_docs


class InstructionParser:

    keyword = "instructions"

    @staticmethod
    def parse(definition_output: DefinitionParserOutput, path):

        specification = definition_output.specification
        symbol_table = definition_output.symbol_table

        if InstructionParser.keyword in specification:
            if len(specification[InstructionParser.keyword].keys()) > 1:
                logging.warning(f"InstructionParser: multiple instructions were listed in the specification (under keys "
                                f"{str(list(specification[InstructionParser.keyword].keys()))[1:-1]}). "
                                "These instructions are independent and results from one instruction are not available to others. "
                                "If this is the intended behavior, please ignore this warning. If the output of one instruction is needed for the "
                                "other, please use separate YAML specifications and separate runs to perform the analysis.")

            for key in specification[InstructionParser.keyword]:
                assert type(specification[InstructionParser.keyword][key]) == dict, f"InstructionParser: instructions are incorrectly defined. Please make sure each instruction is defined under a unique keyword, like this:\n" \
                                                                                    f"instructions:\n" \
                                                                                    f"  my_instruction:\n" \
                                                                                    f"    type: InstructionType\n" \
                                                                                    f"    ... other parameters\n\n" \
                                                                                    f"Found keyword '{key}' containing '{specification[InstructionParser.keyword][key]}', which is a {type(specification[InstructionParser.keyword][key]).__name__}, " \
                                                                                    f"expected a dictionary instead."

                specification[InstructionParser.keyword][key], symbol_table = \
                    InstructionParser.parse_instruction(key, specification[InstructionParser.keyword][key], symbol_table, path)
        else:
            specification[InstructionParser.keyword] = {}

        return symbol_table, specification[InstructionParser.keyword]

    @staticmethod
    @log
    def parse_instruction(key: str, instruction: dict, symbol_table: SymbolTable, path) -> tuple:

        ParameterValidator.assert_keys_present(list(instruction.keys()), ["type"], InstructionParser.__name__, key)
        valid_instructions = [cls[:-6] for cls in ReflectionHandler.discover_classes_by_partial_name("Parser", "dsl/instruction_parsers/")]
        ParameterValidator.assert_in_valid_list(instruction["type"], valid_instructions, "InstructionParser", "type")

        default_params = DefaultParamsLoader.load("instructions/", instruction["type"])
        instruction = {**default_params, **instruction}
        parser = ReflectionHandler.get_class_by_name("{}Parser".format(instruction["type"]), "instruction_parsers/")()
        instruction_object = parser.parse(key, instruction, symbol_table, path)

        symbol_table.add(key, SymbolType.INSTRUCTION, instruction_object)
        return instruction, symbol_table

    @staticmethod
    def generate_docs(path: Path):
        inst_path = PathBuilder.build(path / "instructions")
        instructions = sorted(ReflectionHandler.all_nonabstract_subclasses(Instruction, "Instruction", subdirectory='instructions/'), key=lambda x: x.__name__)

        inst_paths = {}

        for instruction in instructions:
            instruction_name = instruction.__name__[:-11]
            if hasattr(InstructionParser, f"make_{instruction_name.lower()}_docs"):
                fn = getattr(InstructionParser, f"make_{instruction_name.lower()}_docs")
                file_path = fn(inst_path)
            else:
                file_path = InstructionParser.make_docs(instruction, instruction_name, inst_path)

            inst_paths[instruction_name] = file_path

        inst_file_path = inst_path / "instructions.rst"
        with inst_file_path.open('w') as file:
            for key, item in inst_paths.items():
                lines = f"{key}\n---------------------------\n\n.. include:: {os.path.relpath(item, EnvironmentSettings.source_docs_path)}\n\n"
                file.writelines(lines)

    @staticmethod
    def make_docs(instruction, name, path: Path):
        file_path = path / f"{name}.rst"

        with file_path.open("w") as file:
            doc_format = DocumentationFormat(instruction, "", DocumentationFormat.LEVELS[1])
            write_class_docs(doc_format, file)
        return file_path



# quality: gold
import json
from uuid import uuid4

import numpy as np

from ligo.data_model.DatasetItem import DatasetItem
from ligo.data_model.receptor.receptor_sequence.SequenceAnnotation import SequenceAnnotation
from ligo.data_model.receptor.receptor_sequence.SequenceMetadata import SequenceMetadata
from ligo.environment.EnvironmentSettings import EnvironmentSettings
from ligo.environment.SequenceType import SequenceType
from ligo.util.NumpyHelper import NumpyHelper


class ReceptorSequence(DatasetItem):
    FIELDS = {'amino_acid_sequence': str, 'nucleotide_sequence': str, 'identifier': str, 'metadata': dict, 'annotation': dict, 'version': str}
    version = "1"

    nt_to_aa_map = {
        "AAA": "K", "AAC": "N", "AAG": "K", "AAT": "N", "ACA": "T", "ACC": "T", "ACG": "T", 'ACT': 'T',
        'AGA': 'R', "AGC": "S", "AGG": "R", "AGT": "S", "ATA": "I", "ATC": "I", 'ATG': "M", "ATT": "I",
        "CAA": "Q", "CAC": "H", "CAG": "Q", 'CAT': "H", "CCA": "P", "CCC": "P", "CCG": "P", 'CCT': "P",
        "CGA": "R", "CGC": "R", "CGG": "R", "CGT": "R", "CTA": "L", "CTC": "L", "CTG": "L", "CTT": "L",
        "GAA": "E", "GAC": "D", "GAG": "E", 'GAT': "D", "GCA": "A", "GCC": "A", 'GCG': "A", "GCT": "A",
        "GGA": "G", "GGC": "G", "GGG": "G", "GGT": "G", "GTA": "V", "GTC": "V", "GTG": "V", "GTT": "V",
        "TAA": None, "TAC": "Y", "TAG": None, "TAT": "Y", "TCA": "S", "TCC": 'S', "TCG": "S", "TCT": "S",
        "TGA": None, "TGC": "C", "TGG": "W", "TGT": "C", "TTA": "L", "TTC": "F", "TTG": "L", "TTT": "F"
    }

    @classmethod
    def create_from_record(cls, record: np.void):
        if 'version' in record.dtype.names and record['version'] == cls.version:
            return ReceptorSequence(**{**{key: record[key] for key, val_type in ReceptorSequence.FIELDS.items()
                                          if val_type == str and key != 'version'},
                                       **{'metadata': SequenceMetadata(**json.loads(record['metadata'])) if record['metadata'] != '' else None,
                                          'annotation': SequenceAnnotation(**json.loads(record['annotation'])) if record['annotation'] != ''
                                          else None}})
        else:
            raise NotImplementedError

    @classmethod
    def get_record_names(cls):
        return [key for key in cls.FIELDS]

    def __init__(self,
                 amino_acid_sequence: str = None,
                 nucleotide_sequence: str = None,
                 identifier: str = None,
                 annotation: SequenceAnnotation = None,
                 metadata: SequenceMetadata = SequenceMetadata()):
        self.identifier = identifier if identifier is not None and identifier != "" else uuid4().hex
        self.amino_acid_sequence = amino_acid_sequence
        self.nucleotide_sequence = nucleotide_sequence
        self.annotation = annotation
        self.metadata = metadata

    def __repr__(self):
        return f"ReceptorSequence(sequence_aa={self.amino_acid_sequence}, sequence={self.nucleotide_sequence}, " \
               f"annotation={vars(self.annotation) if self.annotation is not None else '{}'}, " \
               f"metadata={vars(self.metadata) if self.metadata is not None else '{}'})"

    def set_metadata(self, metadata: SequenceMetadata):
        self.metadata = metadata

    def set_annotation(self, annotation: SequenceAnnotation):
        self.annotation = annotation

    def get_sequence(self, sequence_type: SequenceType = None):
        """Returns receptor_sequence (nucleotide/amino acid) that corresponds to provided sequence type or preset receptor_sequence type from
        EnvironmentSettings class if no type is provided"""

        sequence_type_ = EnvironmentSettings.get_sequence_type() if sequence_type is None else sequence_type
        if sequence_type_ == SequenceType.AMINO_ACID:
            return self.amino_acid_sequence
        else:
            return self.nucleotide_sequence

    def set_sequence(self, sequence: str, sequence_type: SequenceType):
        if sequence_type == SequenceType.AMINO_ACID:
            self.amino_acid_sequence = sequence
        else:
            self.nucleotide_sequence = sequence
            self.amino_acid_sequence = self._convert_to_aa(sequence)

    def _convert_to_aa(self, nt_sequence: str) -> str:
        return ReceptorSequence.nt_to_aa(nt_sequence)

    @classmethod
    def nt_to_aa(cls, nt_sequence: str):
        kmer_length = 3
        kmers = [nt_sequence[i:i + kmer_length] for i in range(0, len(nt_sequence), kmer_length)]
        return "".join([ReceptorSequence.nt_to_aa_map[kmer] for kmer in kmers])

    def get_record(self):
        """exports the sequence object as a numpy record"""
        return [NumpyHelper.get_numpy_representation(getattr(self, name)) if hasattr(self, name) else getattr(ReceptorSequence, name)
                for name in ReceptorSequence.FIELDS.keys()]

    def get_attribute(self, name: str):
        if hasattr(self, name):
            return getattr(self, name)
        elif hasattr(self.metadata, name):
            return getattr(self.metadata, name)
        elif name in self.metadata.custom_params:
            return self.metadata.custom_params[name]
        elif hasattr(self.annotation, name):
            return getattr(self.annotation, name)
        else:
            raise KeyError(f"ReceptorSequence {self.identifier} does not have attribute {name}.")



import json
from uuid import uuid4

import numpy as np

from ligo.data_model.receptor.Receptor import Receptor
from ligo.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence


class TCGDReceptor(Receptor):
    FIELDS = {'gamma': object, 'delta': object, 'identifier': str, 'metadata': dict, 'version': str}
    version = "1"

    @classmethod
    def create_from_record(cls, record: np.void):
        if 'version' in record.dtype.names and record['version'] == TCGDReceptor.version:

            gamma_record = record[['gamma_' + name for name in ReceptorSequence.get_record_names()]]
            gamma_record.dtype.names = ReceptorSequence.get_record_names()

            delta_record = record[['delta_' + name for name in ReceptorSequence.get_record_names()]]
            delta_record.dtype.names = ReceptorSequence.get_record_names()
            metadata = json.loads(record['metadata']) if record['metadata'] != '' else None

            return TCGDReceptor(gamma=ReceptorSequence.create_from_record(gamma_record),
                                delta=ReceptorSequence.create_from_record(delta_record),
                                identifier=record['identifier'], metadata=metadata)
        else:
            raise NotImplementedError(f"Supported ({TCGDReceptor.version}) and available version differ, but there is no converter available.")

    @classmethod
    def get_record_names(cls):
        return ['gamma_' + name for name in ReceptorSequence.get_record_names()] \
               + ['delta_' + name for name in ReceptorSequence.get_record_names()] \
               + [name for name in cls.FIELDS if name not in ['gamma', 'delta']]

    def __init__(self, gamma: ReceptorSequence = None, delta: ReceptorSequence = None, metadata: dict = None, identifier: str = None):
        super().__init__(metadata=metadata, identifier=identifier)
        self.gamma = gamma
        self.delta = delta

    def get_chains(self):
        return ["gamma", "delta"]
